@@ -21,6 +21,7 @@
 #include "gvfsjobqueryinfo.h"
 #include "gvfsjobenumerate.h"
 #include "gvfsjobseekread.h"
+#include "gvfsjobqueryinforead.h"
 #include "gvfsjobread.h"
 
 typedef GVfsBackendClass GVfsBackendRecentClass;
@@ -200,6 +201,33 @@ recent_backend_seek_on_read (GVfsBackend       *vfs_backend,
   return TRUE;
 }
 
+static void
+recent_backend_query_info_on_read (GVfsBackend           *backend,
+                                   GVfsJobQueryInfoRead  *job,
+                                   GVfsBackendHandle      handle,
+                                   GFileInfo             *info,
+                                   GFileAttributeMatcher *matcher)
+{
+  GError *error = NULL;
+  GFileInfo *real_info;
+
+  real_info = g_file_input_stream_query_info (handle,
+                                              job->attributes,
+                                              G_VFS_JOB (job)->cancellable,
+                                              &error);
+  if (real_info)
+    {
+      g_file_info_copy_into (real_info, info);
+      g_vfs_job_succeeded (G_VFS_JOB (job));
+      g_object_unref (real_info);
+    }
+  else
+    {
+      g_vfs_job_failed_from_error (G_VFS_JOB (job), error);
+      g_error_free (error);
+    }
+}
+
 static gboolean
 recent_backend_close_read (GVfsBackend       *vfs_backend,
                            GVfsJobCloseRead  *job,
@@ -323,7 +351,6 @@ recent_backend_enumerate (GVfsBackend           *vfs_backend,
           g_file_info_set_attribute_mask (info, attribute_matcher);
           recent_backend_add_info (item, info);
           g_vfs_job_enumerate_add_info (job, info);
-          g_object_unref (info);
         }
     }
   g_vfs_job_enumerate_done (job);
@@ -405,10 +432,7 @@ reload_recent_items (GVfsBackendRecent *backend)
       if (!gtk_recent_info_is_local (recent_info)
           || gtk_recent_info_get_private_hint (recent_info)
           || g_strcmp0 (gtk_recent_info_get_mime_type (recent_info), "inode/directory") == 0)
-        {
-          gtk_recent_info_unref (recent_info);
-          continue;
-        }
+        continue;
 
       uri = gtk_recent_info_get_uri (recent_info);
       guid = g_hash_table_lookup (backend->uri_map, uri);
@@ -685,6 +709,7 @@ g_vfs_backend_recent_class_init (GVfsBackendRecentClass *class)
   backend_class->try_open_for_read = recent_backend_open_for_read;
   backend_class->try_read = recent_backend_read;
   backend_class->try_seek_on_read = recent_backend_seek_on_read;
+  backend_class->query_info_on_read = recent_backend_query_info_on_read;
   backend_class->try_close_read = recent_backend_close_read;
   backend_class->try_query_info = recent_backend_query_info;
   backend_class->try_query_fs_info = recent_backend_query_fs_info;
